@@ -3,6 +3,7 @@
 #
 # unit tests for littletable library
 #
+import ast
 from collections import namedtuple
 import io
 import itertools
@@ -19,6 +20,7 @@ import littletable as lt
 try:
     import dataclasses
 except ImportError:
+    print("dataclasses tests disabled")
     # pre Py3.7 (or 3.6 with backported dataclasses)
     dataclasses = None
 else:
@@ -31,6 +33,7 @@ else:
 try:
     import pydantic
 except ImportError:
+    print("pydantic tests disabled")
     pydantic = None
 else:
     class DataPydanticModel(pydantic.BaseModel):
@@ -54,6 +57,35 @@ else:
         b: Optional[Union[int, str]]
         c: Optional[Union[int, str]]
 
+try:
+    import attr
+except ImportError:
+    print("attrs tests disabled")
+    attr = None
+else:
+    AttrClass = attr.make_class("AttrClass", ["a", "b", "c"])
+
+try:
+    import traitlets
+except ImportError:
+    print("traitlets tests disabled")
+    traitlets = None
+else:
+    class TraitletsClass(lt.HasTraitsMixin, traitlets.HasTraits):
+        a = traitlets.Union([traitlets.Int(), traitlets.Unicode()], allow_none=True)
+        b = traitlets.Union([traitlets.Int(), traitlets.Unicode()], allow_none=True)
+        c = traitlets.Union([traitlets.Int(), traitlets.Unicode()], allow_none=True)
+
+        def __init__(self, **kwargs):
+            super().__init__()
+            for k, w in kwargs.items():
+                setattr(self, k, w)
+
+        def __repr__(self):
+            return "{}:(a={}, b={}, c={})".format(type(self).__name__, self.a, self.b, self.c)
+
+        def __dir__(self):
+            return self.trait_names()
 
 DataTuple = namedtuple("DataTuple", "a b c")
 
@@ -66,7 +98,7 @@ except ImportError:
     lt.Table.present = lambda *args, **kwargs: None
 
 
-class Slotted(object):
+class Slotted:
     __slots__ = ['a', 'b', 'c']
 
     def __init__(self, a, b, c):
@@ -166,6 +198,10 @@ def make_test_classes(cls):
         make_test_class(cls, UsingPydanticModel)
         make_test_class(cls, UsingPydanticImmutableModel)
         make_test_class(cls, UsingPydanticORMModel)
+    if attr is not None:
+        make_test_class(cls, UsingAttrClass)
+    if traitlets is not None:
+        make_test_class(cls, UsingTraitletsClass)
 
 
 class AbstractContentTypeFactory:
@@ -214,6 +250,18 @@ else:
     UsingPydanticImmutableModel = AbstractContentTypeFactory
     UsingPydanticORMModel = AbstractContentTypeFactory
 
+if attr is not None:
+    class UsingAttrClass(AbstractContentTypeFactory):
+        data_object_type = AttrClass
+else:
+    UsingAttrClass = AbstractContentTypeFactory
+
+if traitlets is not None:
+    class UsingTraitletsClass(AbstractContentTypeFactory):
+        data_object_type = TraitletsClass
+else:
+    UsingTraitletsClass = AbstractContentTypeFactory
+
 
 def load_table(table, rec_factory_fn, table_size):
     test_size = table_size
@@ -740,12 +788,21 @@ class TableListTests:
                              stat_rec,
                              "invalid stat for {}".format(fieldname))
 
+    def test_stats2(self):
+        self._test_init()
+        field_names = self.t1.info()["fields"]
+        num_fields = len(field_names)
         t1_stats = self.t1.stats(by_field=False)
         for stat, value in (('min', 0), ('max', self.test_size - 1), ('count', self.test_size ** num_fields),):
             for fieldname in field_names:
                 self.assertEqual(value, getattr(t1_stats.by.stat[stat], fieldname),
                                  "invalid {} stat for {}".format(stat, fieldname))
 
+    def test_stats3(self):
+        self._test_init()
+        field_names = self.t1.info()["fields"]
+        num_fields = len(field_names)
+
         # verify that stats can "step over" non-numeric data
         try:
             self.t1[0].a = "not a number"
@@ -764,6 +821,37 @@ class TableListTests:
         t1_stats.present()
         self.assertEqual(self.test_size ** num_fields - 1, t1_stats.by.name["a"].count)
 
+    def test_stats4(self):
+        t1 = lt.Table().csv_import(textwrap.dedent("""\
+        a,b
+        1,2
+        3,
+        5,4
+        """), transforms={}.fromkeys(["a", "b"], int))
+        t1_stats = t1.stats()
+        t1_stats.present()
+        print(t1_stats.info())
+
+        expected = lt.Table().csv_import(textwrap.dedent("""\
+        name,mean,min,max,variance,std_dev,count,missing
+        a,3.0,1,5,4,2.0,3,0
+        b,3.0,2,4,2,1.4142135623730951,2,1
+        """), transforms={}.fromkeys("mean min max variance std_dev count missing".split(), ast.literal_eval))
+        expected.present()
+        print(expected.info())
+
+        self.assertEqual(expected.info()["fields"], t1_stats.info()["fields"])
+
+        for expected_row, row in zip(expected, t1_stats):
+            self.assertEqual(expected_row.name, row.name)
+            self.assertEqual(expected_row.mean, row.mean)
+            self.assertEqual(expected_row.min, row.min)
+            self.assertEqual(expected_row.max, row.max)
+            self.assertEqual(expected_row.variance, row.variance)
+            self.assertEqual(expected_row.std_dev, row.std_dev)
+            self.assertEqual(expected_row.count, row.count)
+            self.assertEqual(expected_row.missing, row.missing)
+
     def test_splitby(self):
         self._test_init()
         is_odd = lambda rec: rec.a % 2
@@ -1117,21 +1205,94 @@ class TableOutputTests:
             """)
         self.assertEqual(expected, out.getvalue())
 
+        # test groupby
+        table = lt.Table().csv_import(textwrap.dedent("""\
+            a,b,default
+            10,100,purple
+            15,150,purple
+            20,200,orange
+            """))
+        table.present()
+        table.present(box=box.ASCII, groupby="default")
+        out = StringIO()
+        table.present(file=out, box=box.ASCII, groupby="default")
+        expected = textwrap.dedent("""\
+            +--------------------+
+            | A  | B   | Default |
+            |----+-----+---------|
+            | 10 | 100 | purple  |
+            | 15 | 150 |         |
+            | 20 | 200 | orange  |
+            +--------------------+
+            """)
+        self.assertEqual(expected, out.getvalue())
+
+        table = lt.Table().csv_import(textwrap.dedent("""\
+            a,b,default
+            10,100,purple
+            15,150,purple
+            15,200,orange
+            15,250,orange
+            20,250,orange
+            """))
+        table.present()
+        table.present(box=box.ASCII, groupby="default a")
+        out = StringIO()
+        table.present(file=out, box=box.ASCII, groupby="default a")
+        expected = textwrap.dedent("""\
+            +--------------------+
+            | A  | B   | Default |
+            |----+-----+---------|
+            | 10 | 100 | purple  |
+            | 15 | 150 |         |
+            | 15 | 200 | orange  |
+            |    | 250 |         |
+            | 20 | 250 |         |
+            +--------------------+
+            """)
+        self.assertEqual(expected, out.getvalue())
+
+        table = lt.Table().csv_import(textwrap.dedent("""\
+            a,b,default
+            10,100,purple
+            15,200,orange
+            15,150,purple
+            20,250,orange
+            15,250,orange
+            """))
+        table.sort("default desc,a")
+        table.present()
+        table.present(box=box.ASCII, groupby="default a")
+        out = StringIO()
+        table.present(file=out, box=box.ASCII, groupby="default a")
+        expected = textwrap.dedent("""\
+            +--------------------+
+            | A  | B   | Default |
+            |----+-----+---------|
+            | 10 | 100 | purple  |
+            | 15 | 150 |         |
+            | 15 | 200 | orange  |
+            |    | 250 |         |
+            | 20 | 250 |         |
+            +--------------------+
+            """)
+        self.assertEqual(expected, out.getvalue())
+
     def test_markdown(self):
         table = lt.Table().csv_import(textwrap.dedent("""\
             a,b
             10,100
             20,200
             """))
-        out = table.as_markdown()
-        print(out)
+        out_markdown = table.as_markdown()
+        print(out_markdown)
         expected = textwrap.dedent("""\
             | a | b |
             |---|---|
             | 10 | 100 |
             | 20 | 200 |
             """)
-        self.assertEqual(expected, out)
+        self.assertEqual(expected, out_markdown)
 
         # test bugfix when table has attribute "default"
         table = lt.Table().csv_import(textwrap.dedent("""\
@@ -1140,8 +1301,8 @@ class TableOutputTests:
             15,150,
             20,200,orange
             """))
-        out = table.as_markdown()
-        print(out)
+        out_markdown = table.as_markdown()
+        print(out_markdown)
         expected = textwrap.dedent("""\
             | a | b | default |
             |---|---|---|
@@ -1149,8 +1310,47 @@ class TableOutputTests:
             | 15 | 150 |  |
             | 20 | 200 | orange |
             """)
-        self.assertEqual(expected, out)
+        self.assertEqual(expected, out_markdown)
 
+        # test grouping in as_markdown
+        table = lt.Table().csv_import(textwrap.dedent("""\
+            a,b,default
+            10,100,purple
+            15,150,purple
+            20,200,orange
+            """))
+        out_markdown = table.as_markdown(groupby="default")
+        print(out_markdown)
+        expected = textwrap.dedent("""\
+            | a | b | default |
+            |---|---|---|
+            | 10 | 100 | purple |
+            | 15 | 150 |  |
+            | 20 | 200 | orange |
+            """)
+        self.assertEqual(expected, out_markdown)
+
+        table = lt.Table().csv_import(textwrap.dedent("""\
+            a,b,default
+            10,100,purple
+            15,200,orange
+            15,150,purple
+            20,250,orange
+            15,250,orange
+            """))
+        table.sort("default desc,a")
+        out_markdown = table.as_markdown(groupby="default a")
+        print(out_markdown)
+        expected = textwrap.dedent("""\
+            | a | b | default |
+            |---|---|---|
+            | 10 | 100 | purple |
+            | 15 | 150 |  |
+            | 15 | 200 | orange |
+            |  | 250 |  |
+            | 20 | 250 |  |
+            """)
+        self.assertEqual(expected, out_markdown)
 
 # sample import data sets
 csv_data = """\
@@ -1321,20 +1521,69 @@ class TableImportExportTests:
             "abc.csv.xz",
         ]
         for name in compressed_files:
-            tt2 = lt.Table().csv_import("test/" + name, transforms=dict.fromkeys("abc", int))
+            import_source_name = "test/" + name
+            tt2 = lt.Table().csv_import(import_source_name, transforms=dict.fromkeys("abc", int))
             print(name, tt2.info())
-            self.assertEqual(tt.info(), tt2.info())
+            expected_info = {**tt.info(), "name": import_source_name}
+            self.assertEqual(expected_info, tt2.info())
             self.assertEqual(sum(tt.all.a), sum(tt2.all.a))
             self.assertEqual(sum(tt.all.b), sum(tt2.all.b))
             self.assertEqual(sum(tt.all.c), sum(tt2.all.c))
 
+        # test separately, no transforms for JSON imports
+        import_source_name = "test/abc.json.gz"
         tt2 = lt.Table().json_import("test/abc.json.gz")
         print("abc.json.gz", tt2.info())
-        self.assertEqual(tt.info(), tt2.info())
+        expected_info = {**tt.info(), "name": import_source_name}
+        self.assertEqual(expected_info, tt2.info())
         self.assertEqual(sum(tt.all.a), sum(tt2.all.a))
         self.assertEqual(sum(tt.all.b), sum(tt2.all.b))
         self.assertEqual(sum(tt.all.c), sum(tt2.all.c))
 
+    def test_csv_import_source_info(self):
+        imports = [
+            ("abc.csv", lt.ImportSourceType.file),
+            ("abc.tsv", lt.ImportSourceType.file),
+            ("abc.csv.zip", lt.ImportSourceType.zip),
+            ("abc.csv.gz", lt.ImportSourceType.gzip),
+            ("abc.csv.xz", lt.ImportSourceType.lzma),
+            ("a,b,c\n1,2,3", lt.ImportSourceType.string),
+        ]
+        for fname, expected_type in imports:
+            if "\n" not in fname:
+                import_name = "test/" + fname
+            else:
+                import_name = fname
+            if not import_name.endswith(".tsv"):
+                tbl = lt.Table().csv_import(import_name)
+            else:
+                tbl = lt.Table().tsv_import(import_name)
+
+            print(repr(import_name), tbl.import_source, tbl.import_source_type)
+
+            if "\n" not in fname:
+                self.assertEqual(import_name, tbl.import_source)
+            else:
+                self.assertEqual(None, tbl.import_source)
+            self.assertEqual(expected_type, tbl.import_source_type)
+
+    def test_csv_import_from_url(self):
+        import subprocess
+        import urllib.request
+
+        web_address = "http://localhost:8888"
+        web_server = subprocess.Popen("python test/csv_import_http_server.py".split())
+        url = web_address + "/abc.csv"
+        tbl = lt.Table().csv_import(url)
+        tbl.present()
+        self.assertEqual(url, tbl.import_source)
+        self.assertEqual(lt.ImportSourceType.url, tbl.import_source_type)
+
+        with urllib.request.urlopen(web_address + "/EXIT"):
+            pass
+
+        web_server.wait()
+
     def test_csv_filtered_import(self):
         test_size = 3
         tt = lt.Table().csv_import("test/abc.csv", transforms=dict.fromkeys("abc", int))
